
// jwc 2016-0901-1300 http://caolan.github.io/async/docs.html 
// * Discover Xof2
//
// +  Store Info for Print Later
// ++ Print Stored Info: Xof3

var async = require('async');
var noble = require('../index');

var peripheralIdOrAddress = process.argv[2].toLowerCase();

  noble.on('stateChange', function(state) {
    if (state === 'poweredOn') {
      noble.startScanning();
    } else {
      noble.stopScanning();
    }
  });

  // Discover/Search specifically for 'peripheralIdOrAddress' 
  //   - to print out about 'peripheral.advertisement[.localName:'Light Switch', .serviceData:'', .serviceUuids;'ff10']
  noble.on('discover', function(peripheral) {
    if (peripheral.id === peripheralIdOrAddress || peripheral.address === peripheralIdOrAddress) {
      noble.stopScanning();

      // print peripheral id that was found: 'fc12b80f1c60;  
      console.log('peripheral with ID ' + peripheral.id + ' found');
      
      
      // [copied from above] - to print out about 'peripheral.advertisement[.localName:'Light Switch', .serviceData:'', .serviceUuids;'ff10']
      //  
      var advertisement = peripheral.advertisement;

      var localName = advertisement.localName;
      var txPowerLevel = advertisement.txPowerLevel;
      var manufacturerData = advertisement.manufacturerData;
      var serviceData = advertisement.serviceData;
      var serviceUuids = advertisement.serviceUuids;

      if (localName) {
        console.log('  Local Name        = ' + localName);
      }

      if (txPowerLevel) {
        console.log('  TX Power Level    = ' + txPowerLevel);
      }

      if (manufacturerData) {
        console.log('  Manufacturer Data = ' + manufacturerData.toString('hex'));
      }

      if (serviceData) {
        console.log('  Service Data      = ' + serviceData);
      }

      if (serviceUuids) {
        console.log('  Service UUIDs     = ' + serviceUuids);
      }

      console.log();

      // move on to next function/stage
      explore(peripheral);
    }
  });

  function explore(peripheral) {
    
    // ++ Print 1of3: Services and Their Characteristics: Print Report Title: 'services and characteristics:'
    console.log('services and characteristics:');

    peripheral.on('disconnect', function() {
      // jwc 2016-0901-1545
      console.log('>>> Peripheral Disconnect: 2of2');
      
      process.exit(0);
    });
    
    peripheral.connect(function(error) {
      
      // * Discover 1of2: Services: 'services'
      //
      // Find & Print All Services: ['FF10'] (only one for now) via upcoming 'async.whilst: 1 of 2'
      //
      peripheral.discoverServices([], function(error, services) {
        var serviceIndex = 0;

        // async.whilst: 1 of 2: * Discover Services: 'services'
        async.whilst(
        
          // async.whilst: Arg-1of3: condition-test: All Services
          function () {
            return (serviceIndex < services.length);
          },
          
          // async.whilst: Arg-2of3: iteratee-on-condition-test-success
          function(callback) {
            
            // 'serviceIndex' undergo 'serviceIndex++' only after iterate through all its own characteristics (at the end of 'async.whilst: 2 of 2:')
            var service = services[serviceIndex];
            var serviceInfo = service.uuid;

            if (service.name) {
              serviceInfo += ' (' + service.name + ')';
            }
            
            // ++ Print 2of3: Services and Their Characteristics: ALL Services:['1800 (Generic Access)', '1801 (Generic Attribute)', 'ff10']
            //   - under each above service, will print at next sub-level its own characteristics (next nested loop)
            console.log(serviceInfo);

            
            // * Discover Characteristics: 'characteristics'
            //
            // Find & Print All Characteristics for each above Service via upcoming 'async.whilst: 2 of 2'
            //   - 'service.uuid (service.name)'  [Level-1: processed from above]
            //      - 'characteristic.uuid (characteristic.name)' [Level-2: processed from below]
            //
            //   - '1800 (Generic Access)'
            //      - '2a00 (Device Name)'
            //   - '1801 (Generic Attribute)'
            //      - '2a01 (Appearance)'
            //   - 'ff10'
            //      - 'FF11 (Switch)'
            //      - 'FF12 (State)'
            //
            service.discoverCharacteristics([], function(error, characteristics) {
              var characteristicIndex = 0;

              // async.whilst: 2 of 2: * Discover Characteristics: 'characteristics'
              async.whilst(
              
                // async.whilst: Arg-1of3: condition-test
                function () {
                  return (characteristicIndex < characteristics.length);
                },
                
                // async.whilst: Arg-2of3: iteratee-on-condition-test-success
                function(callback) {
                  var characteristic = characteristics[characteristicIndex];
                  //
                  // + Store Characteric:Header 1of2: 'characteristic.uuid'
                  var characteristicInfo = '  ' + characteristic.uuid;
                  //
                  // + Store Characteric:Header 2of2: 'characteristic.name'
                  if (characteristic.name) {
                    characteristicInfo += ' (' + characteristic.name + ')';
                  }

                  // async.series (forced-sequentially through entire list)
                  async.series([
                  
                    // * Discover 2of2: Descriptors: 'descriptors'
                    //
                    // async.series list of functions: Task-1of3: Discover/Search forced-sequentially for each 'characteristic.discoverDescriptors'
                    //
                    function(callback) {
                      characteristic.discoverDescriptors(function(error, descriptors) {
                                               
                        // async.detect: for randomly first-found iteratee from 'descriptors'
                        //   - actually safe for this 'async.detect' which discovers/searches 'randomly', since there is only one descriptor per characteristic anyways
                        //     -  thus no multiple descriptors to 'async.series'(sequentially) traverse through
                        async.detect(
                        
                          // async.detect: Arg-1of3: collection in 'descriptors' to iterate over
                          //   - as defined by above 'characteristic.discoverDescriptors(function(error, descriptors)'
                          //   - and as mentioned above, each characteristic should only have one descriptor (also one value)
                          descriptors,  
                          
                          // async.detect: Arg-2of3: condition-test for randomly first-found iteratee
                          // all descriptors uuid should == '2901'
                          function(descriptor, callback) {
                            return callback(descriptor.uuid === '2901');
                          },
                          
                          // async.detect: Arg-3of3: Callback on randomly first-found iteratee
                          //   - does not seem to be used in current example 'Light Switch', thus no appending to 'characteristicInfo'
                          //   - since no manual user override by default, but could be done during program running by manual user
                          //     - and if manually used, then will print at same level of Characteristic:Header
                          function(userDescriptionDescriptor){
                            if (userDescriptionDescriptor) {
                              userDescriptionDescriptor.readValue(function(error, data) {
                                if (data) {
                                  characteristicInfo += ' (' + data.toString() + ')';
				}else{
                                  characteristicInfo += ' ( temptest ) ';
				}
                                callback();
                              });
                            } else {
                                callback();
                            }
                          }
                          
                          
                        );
                      });
                      
                    },
                    
                    // async.series list of functions: Task-2of3: Get Characteristic's Content: Properties & Value
                    function(callback) {
                      // + Store Characteristic:Property 
                      // begin with create new line '\n'
                      characteristicInfo += '\n    properties  ' + characteristic.properties.join(', ');

                      // check to insure 'read' is present and not missing (-1) to insure this is a value to print
                      if (characteristic.properties.indexOf('read') !== -1) {
                        characteristic.read(function(error, data) {
                          
                          // + Store Characteristic:Value
                          if (data) {
                            
                            // with current example 'Light Swtich', 'string' sometimes blank/not-used except for Characteristic: '2a00': 'Smart Light Switch'
                            var string = data.toString('ascii');

                            // begin with create new line '\n'
                            characteristicInfo += '\n    value       ' + data.toString('hex') + ' | \'' + string + '\'';
                          }
                          callback();
                        });
                      } else {
                        callback();
                      }
                      
                    },
                    
                    // async.series list of functions: Task-3of3: Print
                    // ++ Print 3of3: Services and Their Characteristics: Print ALL Characteristic:[Property, Value]
                    function() {
                      // print out 'characteristicInfo'
                      console.log(characteristicInfo);
                      characteristicIndex++;
                      callback();
                    }
                  ]);
                  
                },
                
                // async.whilst: Arg-3of3: callback-on-condition-test-failure
                //   - [copied from above]'serviceIndex' undergo 'serviceIndex++' only after interate through all its own characteristics (at the end of 'async.whilst: 2 of 2:')
                function(error) {
                  serviceIndex++;
                  callback();
                }
              );
            });
          },

          // async.whilst: Arg-3of3: callback-on-condition-test-failure
          function (err) {
            peripheral.disconnect();
            console.log('>>> Peripheral Disconnect: 1of2');
          }
          
        );  // async.whilst: 1 of 2
      });  // peripheral.discoverServices([], function(error, services) {
      
      
    });
  }

