
// jwc 2016-0901-1300 http://caolan.github.io/async/docs.html 
// * Discover Xof2
//
// +  Store Info for Print Later
// ++ Print Stored Info: Xof3


/* 
    Input: 
    pi@raspberrypi:~/Jwc-2016-0728-2345/BluetoothToArduino/04-Sandeepmistry-Noble/noble/examples $ sudo node peripheral-explorer-Formatted-2016-0901-0330e.js FC:12:B8:0F:1C:60
    
    Output:
    peripheral with ID fc12b80f1c60 found
      Local Name        = Light Switch
      Service Data      =
      Service UUIDs     = ff10

    services and characteristics:
    1800 (Generic Access)
      2a00 (Device Name)
        properties  read
        value       536d617274204c6967687420537769746368 | 'Smart Light Switch'
      2a01 (Appearance)
        properties  read
        value       0000 | ''
    1801 (Generic Attribute)
      2a05 (Service Changed)
        properties  indicate
    ff10
      ff11 (Switch)
        properties  read, write
        value       00 | ''
      ff12 (State)
        properties  notify
    >>> Peripheral Disconnect: 1of2
    >>> Peripheral Disconnect: 2of2
 */

var async = require('async');
var noble = require('../index');

var peripheralIdOrAddress = process.argv[2].toLowerCase();

  noble.on('stateChange', function(state) {
    if (state === 'poweredOn') {
      noble.startScanning();
      console.log('1a> noble.on(stateChange) >> noble.startScanning()');

    } else {
      noble.stopScanning();
      console.log('1b> noble.on(stateChange) >> noble.stopScanning()');

    }
  });

  // Discover/Search specifically for 'peripheralIdOrAddress' 
  //   - to print out about 'peripheral.advertisement[.localName:'Light Switch', .serviceData:'', .serviceUuids;'ff10']
  noble.on('discover', function(peripheral) {
    if (peripheral.id === peripheralIdOrAddress || peripheral.address === peripheralIdOrAddress) {
      noble.stopScanning();
      
      console.log('2> noble.on(discover) >> (peripheral.id === peripheralIdOrAddress || peripheral.address === peripheralIdOrAddress) >> noble.stopScanning()');

      // print peripheral id that was found: 'fc12b80f1c60;  
      console.log('peripheral with ID ' + peripheral.id + ' found');
      
      
      // [copied from above] - to print out about 'peripheral.advertisement[.localName:'Light Switch', .serviceData:'', .serviceUuids;'ff10']
      //  
      var advertisement = peripheral.advertisement;

      var localName = advertisement.localName;
      var txPowerLevel = advertisement.txPowerLevel;
      var manufacturerData = advertisement.manufacturerData;
      var serviceData = advertisement.serviceData;
      var serviceUuids = advertisement.serviceUuids;

      if (localName) {
        console.log('  Local Name        = ' + localName);
      }

      if (txPowerLevel) {
        console.log('  TX Power Level    = ' + txPowerLevel);
      }

      if (manufacturerData) {
        console.log('  Manufacturer Data = ' + manufacturerData.toString('hex'));
      }

      if (serviceData) {
        console.log('  Service Data      = ' + serviceData);
      }

      if (serviceUuids) {
        console.log('  Service UUIDs     = ' + serviceUuids);
      }

      console.log();

      // move on to next function/stage
      explore(peripheral);
    }
  });

  function explore(peripheral) {
    
    console.log('3> function explore(peripheral)');

    // ++ Print 1of3: Services and Their Characteristics: Print Report Title: 'services and characteristics:'
    console.log('services and characteristics:');

    peripheral.on('disconnect', function() {
      // jwc 2016-0901-1545
      console.log('>>> Peripheral Disconnect: 2of2');
      
      process.exit(0);
    });
    
    peripheral.connect(function(error) {
      
      console.log('3-1> function explore(peripheral) >> peripheral.connect()');

      // * Discover 1of4: Services: 'services'
      //
      // Find & Print All Services: ['FF10'] (only one for now) via upcoming 'async.whilst: 1 of 2'
      //
      // ** Specificly 'ff10'
      //
      peripheral.discoverServices(['ff10'], function(error, services) {
        
        console.log('3-1-1> function explore(peripheral) >> peripheral.connect() >> peripheral.discoverServices([\'ff10\']...');

        var serviceIndex = 0;

        // async.whilst-1of2: * Discover Services: 'services'
        async.whilst(
        
          // async.whilst-1of2: Arg-1of3: condition-test: All Services
          function () {
            return (serviceIndex < services.length);
          },
          
          // async.whilst-1of2: Arg-2of3: iteratee-on-condition-test-success
          function(callback) {
            
            // 'serviceIndex' undergo 'serviceIndex++' only after iterate through all its own characteristics (at the end of 'async.whilst: 2 of 2:')
            var service = services[serviceIndex];
            var serviceInfo = service.uuid;

            if (service.name) {
              serviceInfo += ' (' + service.name + ')';
            }
            
            // ++ Print 2of3: Services and Their Characteristics: ALL Services:['1800 (Generic Access)', '1801 (Generic Attribute)', 'ff10']
            //   - under each above service, will print at next sub-level its own characteristics (next nested loop)
            console.log(serviceInfo);

            
            // * Discover 2of4: Characteristics: 'characteristics'
            //
            /* 

            * Find & Print All Characteristics for each above Service via upcoming 'async.whilst: 2 of 2'
                
                service.uuid (service.name)  [Level-1: processed from above]
                   characteristic.uuid (characteristic.name) [Level-2: processed from below]
                      characteristic.properties [Level-3: processed from below]
                      characteristic.read iff 'characteristic.properties=read' [Level-3 optional: processed from below]
            
                services and characteristics:
                1800 (Generic Access)
                   2a00 (Device Name)
                      properties  read
                      value       536d617274204c6967687420537769746368 | 'Smart Light Switch'
                   2a01 (Appearance)
                      properties  read
                      value       0000 | ''
                1801 (Generic Attribute)
                   2a05 (Service Changed)
                      properties  indicate
                ff10
                   ff11 (Switch)
                      properties  read, write
                      value       00 | ''
                   ff12 (State)
                      properties  notify 
                    
            */

              // ** Specificly 'ff11'
              //
              service.discoverCharacteristics(['ff11'], function(error, characteristics) {
              var characteristicIndex = 0;

              // async.whilst-2of2: * Discover Characteristics: 'characteristics'
              async.whilst(
              
                // async.whilst-2of2: Arg-1of3: condition-test
                function () {
                  return (characteristicIndex < characteristics.length);
                },
                
                // async.whilst-2of2: Arg-2of3: iteratee-on-condition-test-success
                function(callback) {
                  var characteristic = characteristics[characteristicIndex];
                  //
                  // + Store Characteric:Header 1of2: 'characteristic.uuid'
                  var characteristicInfo = '  ' + characteristic.uuid;
                  //
                  // + Store Characteric:Header 2of2: 'characteristic.name'
                  if (characteristic.name) {
                    characteristicInfo += ' (' + characteristic.name + ')';
// characteristic.write(new Buffer([0x01]), false, function(error) {});
// characteristic.write(new Buffer([0x00]), false, function(error) {});
// console.log('1>> ' + characteristic.uuid + ' >> ' + characteristic.name);
                  }

                  // async.series (forced-sequentially through entire list)
                  async.series([
                  
                    // * Discover 3of4: Descriptors: 'descriptors'
                    //
                    // async.series list of functions: Task-1of3: Discover/Search forced-sequentially for each 'characteristic.discoverDescriptors'
                    //
                    function(callback) {
                      characteristic.discoverDescriptors(function(error, descriptors) {
                                               
                        // async.detect: for randomly first-found iteratee from 'descriptors'
                        //   - actually safe for this 'async.detect' which discovers/searches 'randomly', since there is only one descriptor per characteristic anyways
                        //     -  thus no multiple descriptors to 'async.series'(sequentially) traverse through
                        async.detect(
                        
                          // async.detect: Arg-1of3: collection in 'descriptors' to iterate over
                          //   - as defined by above 'characteristic.discoverDescriptors(function(error, descriptors)'
                          //   - and as mentioned above, each characteristic should only have one descriptor (also one value)
                          descriptors,  
                          
                          // async.detect: Arg-2of3: condition-test for randomly first-found iteratee
                          // all descriptors uuid should == '2901'
                          function(descriptor, callback) {
                            return callback(descriptor.uuid === '2901');
                          },
                          
                          // async.detect: Arg-3of3: Callback on randomly first-found iteratee
                          //   - does not seem to be used in current example 'Light Switch', thus no appending to 'characteristicInfo'
                          //   - since no manual user override by default, but could be done during program running by manual user
                          //     - and if manually used, then will print at same level of Characteristic:Header
                          function(userDescriptionDescriptor){
                            //// good temp test for learning this code: characteristicInfo += ' ( userDescriptionDescriptor-test )';
                            if (userDescriptionDescriptor) {
                              userDescriptionDescriptor.readValue(function(error, data) {
                                if (data) {
                                  characteristicInfo += ' (' + data.toString() + ')';
                                }
                                callback();
                              });
                            } else {
                                callback();
                            }
                          }
                          
                          
                        );
                      });
                      
                    },
                    
                    // * Discover 4of4: Characteristics: Properties & Values:
                    //
                    // async.series list of functions: Task-2of3: Get Characteristic's Content: Properties & Value
                    function(callback) {
                      // + Store Characteristic:Property 
                      // begin with create new line '\n'
                      characteristicInfo += '\n    properties  ' + characteristic.properties.join(', ');

                      // check to insure 'read' is present and not missing (-1) to insure this is a value to print
                      if (characteristic.properties.indexOf('read') !== -1) {
                        characteristic.read(function(error, data) {
                          // + Store Characteristic:Value
                          if (data) {
                            
                            // with current example 'Light Swtich', 'string' sometimes blank/not-used except for Characteristic: '2a00': 'Smart Light Switch'
                            var string = data.toString('ascii');

                            // begin with create new line '\n'
                            characteristicInfo += '\n    value       ' + data.toString('hex') + ' | \'' + string + '\'';

                          }
                          callback();
                          });
                        } else {
                            callback();
                        }
                      
                    },



// * Discover 5of4: Characteristics: Properties & Values:
                    //
                    // async.series list of functions: Task-2of3: Get Characteristic's Content: Properties & Value
                    function(callback) {
                      if (characteristic.properties.indexOf('write') !== -1) {
                        characteristic.write(new Buffer([0x01]), false, function(error) { callback(); });
                      } else {
                        callback();
                      }
                    },

// * Discover 6of4: Characteristics: Properties & Values:
                    //
                    // async.series list of functions: Task-2of3: Get Characteristic's Content: Properties & Value
                    function(callback) {
                      if (characteristic.properties.indexOf('write') !== -1) {
                        characteristic.write(new Buffer([0x00]), false, function(error) { callback(); });
                      } else {
                        callback();
                      }
                    },

                    
                    // async.series list of functions: Task-3of3: Print
                    // ++ Print 3of3: Services and Their Characteristics: Print ALL Characteristic:[Property, Value]
                    function() {
                      // print out 'characteristicInfo'
                      console.log(characteristicInfo);
                      
test_2(peripheral ,characteristic);                      
                      
                      characteristicIndex++;
                      callback();
                    }
                            
                  ]);
                  
                },
                
                // async.whilst-2of2: Arg-3of3: callback-on-condition-test-failure
                //   - [copied from above]'serviceIndex' undergo 'serviceIndex++' only after interate through all its own characteristics (at the end of 'async.whilst: 2 of 2:')
                function(error) {
                  serviceIndex++;
                  callback();
                }
              );
            });
          },

          // async.whilst-1of2: Arg-3of3: callback-on-condition-test-failure
          function (err) {
            
test_1(peripheral);
            
            /// archive-good: peripheral.disconnect();
            /// archive-good: console.log('>>> Peripheral Disconnect: 1of2');
          }
          
        );  // async.whilst-1of2
      });  // peripheral.discoverServices([], function(error, services) {
      
      
    });  // peripheral.connect(function(error) {
  }  // function explore(peripheral) {

  
  var alphabetString = 'abcdefghijklmnopqrstuvwxyz';
  
  function test_1(peripheral){
    
    console.log( '> test_1' );
 
 
    for( var i = 0; i <= 25 ; i++ ){
      // //characteristic_GLOBAL.write(new Buffer([0x01]), false, function(error) {});
      // //characteristic_GLOBAL.write(new Buffer([0x00]), false, function(error) {});
      // // characteristic_GLOBAL.write(new Buffer([0x66]), false, function(error) {});
      // // characteristic_GLOBAL.write(new Buffer([0x67]), false, function(error) {});
      characteristic_GLOBAL.write(new Buffer([alphabetString.charCodeAt(i)]), false, function(error) {});
      // // characteristic_GLOBAL.write(new Buffer([alphabetString.charCodeAt(0)]), false, function(error) {});
      console.log( '>> test_1: ' + i );

    }

    // good: console.log('>>> Peripheral Disconnect: 1of2');
    // good: peripheral.disconnect();
    
    clearActive();

  }

  var  peripheral_GLOBAL 
  var  characteristic_GLOBAL

  function test_2(peripheral ,characteristic){
    
    console.log( '> test_2' );

    characteristic.write(new Buffer([0x68]), false, function(error) {});
    characteristic.write(new Buffer([0x69]), false, function(error) {});
    
    peripheral_GLOBAL = peripheral;
    characteristic_GLOBAL = characteristic;
    // peripheral.disconnect();
    // console.log('>>> Peripheral Disconnect: 1of2');

  }
  
  var active = true;

  function clearActive() {
    active = false;
  }
  
  function cooldown(){
    // 100 msec cooldown countdown for next key press
    setTimeout( clearActive, 100 );  
    
  }
    
  const readline = require('readline');
  readline.emitKeypressEvents(process.stdin);
  
  // // process.stdin.setRawMode(true);
  if (process.stdin.setRawMode){
    process.stdin.setRawMode(true);
  } else{
    require('tty').setRawMode(true);
  }
  
  process.stdin.on('keypress', function(str, key) {
    if (key.ctrl && key.name === 'c') {
      // // obsolete: process.exit();
      console.log('>>> Peripheral Disconnect: 1of2');
      peripheral_GLOBAL.disconnect();

    } else if( !active && key ){
      
      console.log(`You pressed the "${str}" key`);
      // // detail info not needed: console.log(key);
      
      if (key.name === 'a') {
        
        console.log('> a');
        for( var i = 0; i <= 0 ; i++ ){
          // // characteristic_GLOBAL.write(new Buffer(['a']), false, function(error) {});
          characteristic_GLOBAL.write(new Buffer([key.name.charCodeAt(0)]), false, function(error) {});
          console.log( '>> for-loop: ' + i );
        }        
        cooldown();
      } else if (key.name === 'b') {
        
        console.log('> b');
        for( var i = 0; i <= 1; i++ ){
          characteristic_GLOBAL.write(new Buffer([key.name.charCodeAt(0)]), false, function(error) {});
          console.log( '>> for-loop: ' + i );
        }        
        cooldown();
      } else if (key.name === 'c') {
        
        console.log('> c');
        for( var i = 0; i <= 2 ; i++ ){
          characteristic_GLOBAL.write(new Buffer([key.name.charCodeAt(0)]), false, function(error) {});
          console.log( '>> for-loop: ' + i );
        }        
        cooldown();      
      }
      console.log();
      
    }
  });
  
  console.log('Ready for any key...');